"""
POEM Synchronizer will try to contact all POEM Web instances configured in
POEM_SYNC_URLS and synchronize all their profiles to the local database. Each
POEM Web instance is represented by a namespace, which needs to be unique. The
following rules apply:
 * Two different urls can't provide the same namespace
 * Configured (ns, url) tuples have always preference over the already existing tuples

Synchronization is performed in a sequence:
 * syncs core objects (metrics, fqans, etc.),
 * sync profiles
 * sync metric instances and establish foreign keys to ATP database objects
 * delete (profile, metric instances) mappings that are no longer used
 * unlink profiles
 
Profiles, metrics, flavours, FQANs and VOs all have isdeleted flag and thus can 
recover IDs once something is deleted and reinserted (this is important for MRS 
and ACE who have FKs to POEM-SYNC tables). 
"""
from itertools import groupby
from urlparse import urljoin
import signal
import re

from django.core.exceptions import ObjectDoesNotExist
from django.db import connection, transaction
from django.core import serializers
from django.utils import simplejson
from django.utils.log import NullHandler

import logging, urllib
import Poem_sync.django_logging
import Poem_sync.settings

from Poem_sync.poem_sync.models import *

LOG_INSTANCE = logging.getLogger('POEM')

class SyncException(Exception):
    
    def __init__(self, value):
        self.parameter = value
        
    def __str__(self):
        return repr(self.parameter)

class PoemSynchronizer(object):
    
    def __init__(self, url=""):
        self._base_url = url
        self._namespace_ob = None
        self._has_error = False
        self._raise_exception = True
        self._profile_exist_list = []
        self._error_count = 0
        self._warn_count = 0
        self._urlopen_timeout = int(Poem_sync.settings.URLOPEN_TIMEOUT)
        self._m_index = dict(MetricMapping.objects.all().values_list('new_metric', 'old_metric'))
        signal.signal(signal.SIGALRM, self.urlopen_handler)
#
#        SPECIAL NOTE 
#        These Lines are added to make Case insensitive searched on ORACLE DB. 
#        MySQL does case insensitive erach on text fields but Oracle does a Case Sensitive search
#        To mitigate these two different behaviour these lines are added. Both the DB now behave 
#        in Case insensitive manner
#        
#        if settings.DATABASE_ENGINE == 'oracle':
#            cursor = connection.cursor()
#            cursor.execute('ALTER SESSION set NLS_COMP=ANSI;')
#            cursor.execute('ALTER SESSION set NLS_SORT=BINARY_CI;')

    def get_error_stat(self):
        return self._error_count

    def get_warn_stat(self):
        return self._warn_count
       
    def _report_error(self, msg_str):
        self._error_count += 1
        LOG_INSTANCE.error(msg_str)
        
    def _report_warning(self, msg_str):
        self._warn_count += 1
        LOG_INSTANCE.warning(msg_str)
        
    def _raise_error(self, error_str):
        self._has_error = True
        msg_str = "SyncException: [%s] : %s" % (self._namespace_ob, error_str)
        self._report_error(msg_str)
        if self._raise_exception:
            raise SyncException(msg_str)
        else:
            return None
                     
    def set_base_url(self, url):
        self._base_url = url
    
    def get_base_url(self):
        return self._base_url
    
    def set_namespace(self, namespace):
        self._namespace_ob = namespace
        self._base_url = namespace.api_url
    
    def get_namespace(self):
        return self._namespace_ob
         
    def update_obj(self, obj, dict_ob):
        for key, val in dict_ob.items():
            if hasattr(obj, key): 
                setattr(obj, key, val)
    
    def sync_metricinstance(self, metricinstance_dict):
        """ Syncs a MetricInstance object """
        metric = self.sync_metric(metricinstance_dict['metric'])
        fqan = self.sync_fqan(metricinstance_dict['fqan'])
        sfla = self.sync_service_type_flavour(metricinstance_dict['atp_service_type_flavour'])
        atp_vo = self.sync_vo(metricinstance_dict['atp_vo'])
        if sfla:
            obj, dummy = MetricInstance.objects.get_or_create(metric=metric, 
                                                              fqan=fqan, 
                                                              atp_service_type_flavour=sfla, 
                                                              atp_vo=atp_vo)
            return obj
        else:
            return self._raise_error("Unable to sync MetricInstance object due to serviceTypeFlavour")    
        
    def sync_metric(self, metric_name):
        """ Syncs a Metric object """
        try:
            if metric_name in self._m_index.keys():
                m_object = Metric.objects.filter(name=self._m_index[metric_name])
                # replace old metric name with a new one keeping ID 
                if m_object:
                    m_object[0].name = metric_name
                    m_object[0].save()
            
            if metric_name.strip():
                metric, dummy = Metric.objects.get_or_create(name=metric_name)
            else:
                return None
        except Exception, (ex):
            return self._raise_error('From sync_metric: %s %s' % (metric_name, ex))
        return metric
     
    def sync_fqan(self, fqan):
        """ Syncs a Fqan object """
        try:
            if not fqan:
                return None
            obj, dummy = Fqan.objects.get_or_create(fqan=fqan)
        except Exception, (ex):
            return self._raise_error('From sync_fqan: %s %s' % (fqan, ex))
        return obj
     
    def sync_service_type_flavour(self, flavourname):
        """ Syncs a ServiceTypeFlavour object """
        try:
            obj = ServiceTypeFlavour.objects.get(flavourname=flavourname)
        except Exception, (ex):
            return self._raise_error('From sync_service_type_flavour: %s %s' % (flavourname, ex))
        return obj
    
    def sync_vo(self, voname):
        """ Syncs a ServiceTypeFlavour object """
        try:
            voname=voname.lower()
            obj = Vo.objects.get(voname=voname)
        except Exception, (ex):
            return self._raise_error('From sync_vo: %s %s' % (voname, ex))
        return obj
    
    def sync_groups(self, groups):
        """ Syncs a ServiceTypeFlavour object """
        try:
            obj = Groups.objects.get(
                    group_type = GroupType.objects.get(typename=groups['atp_grouptype']),
                    groupname = groups['atp_groupname']
                )
        except Exception, (ex):
            return self._raise_error('From sync_groups: %s %s' % (groups, ex))
        return obj
    
    def sync_expression(self, expr_dict):
        """ Syncs an Expression Object """
        try:
            if not (expr_dict['expr'].strip() and expr_dict['expr_type'].strip()):
                return None
            obj, dummy = Expression.objects.get_or_create(name=expr_dict['name'])
            del expr_dict['extension_ptr']
            self.update_obj(obj, expr_dict)
            obj.save()
        except Exception, (ex):
            return self._raise_error('From sync_expression: %s %s' % (expr_dict, ex))
        return obj
        
    def sync_extension(self, name):
        """ Syncs an Expression Object """
        try:
            obj, dummy = Extension.objects.get_or_create(name=name)
        except Exception, (ex):
            return self._raise_error('From sync_extension: %s %s' % (name, ex))
        return obj
    
    def sync_profile(self, p_dict):
        """ Sync A profile Object and all its components """
        # clear the Error flag
        self._has_error = False
        #Check if the namespace is set or not 
        if not self._namespace_ob:
            return self._raise_error("Namespace is not set")
        # check for selctive_profile_sync
        if Poem_sync.settings.NS_RESTRICT and self.selective_profile_list and not self.selective_profile_list.count(p_dict['name']):
            LOG_INSTANCE.info("Ignoring Sync of profile %s" % (p_dict['name']))
            return None
        try:
            pobj1 = Profile(namespace=self._namespace_ob, 
                    name=p_dict['name'], 
                    version=p_dict['version']
                    )
            try:
                pobj = Profile.objects.get(namespace = self._namespace_ob, name=p_dict['name'], version=p_dict['version'])
                self._profile_exist_list.append(pobj)
            except Profile.DoesNotExist:
                pobj = None
            
            pobj1.owner = p_dict['owner']
            pobj1.description = p_dict['description']
            pobj1.valid_from = p_dict['valid_from']
            pobj1.valid_to = p_dict ['valid_to']
            pobj1.isdeleted = 'N'
            # backward compatible with Update-17
            if 'formatted_tags' in p_dict.keys():
                pobj1.tags = p_dict['formatted_tags']
            else:
                pobj1.tags = ''
            pobj1.atp_vo = self.sync_vo(p_dict['atp_vo'])          
            nextensions = self.sync_extension_list(p_dict['extensions'])
            LOG_INSTANCE.info("Adding defining VO as default running for the metrics in the profile")
            minsts = []
            for minst in p_dict['metric_instances']:
                minst['atp_vo'] = p_dict['atp_vo']
                minsts.append(minst)
            p_dict['metric_instances'] = minsts
            LOG_INSTANCE.info("Added '%s' as default running for the metrics in the profile" % (p_dict['atp_vo']))
            nmetricinstances = self.sync_object_list(p_dict['metric_instances'], self.sync_metricinstance)
            ngroups = self.sync_object_list(p_dict['groups'], self.sync_groups)
            
            if self._has_error:
                raise SyncException('Unable to Sync')
        except ValueError, (ve):
            return self._raise_error("From sync_profile ValueError: (%s): %s" % (p_dict['name'], ve))
        except KeyError, (ke):
            return self._raise_error("From sync_profile KeyError: (%s): %s" % (p_dict['name'], ke))  
        except SyncException, (ex):
            return self._raise_error('From sync_profile: (%s): %s' % (p_dict['name'], ex))

        try: 
            pobj = Profile.objects.get(namespace = self._namespace_ob, name=p_dict['name'], version=p_dict['version'])
            pobj.owner = pobj1.owner
            pobj.description = pobj1.description
            pobj.valid_from = pobj1.valid_from
            pobj.valid_to = pobj1.valid_to
            pobj.isdeleted = pobj1.isdeleted
            pobj.atp_vo = pobj1.atp_vo       
            pobj.tags = pobj1.tags
        except Profile.DoesNotExist:
            pobj = pobj1

        try:
            pobj.save()
        except Exception, (ex):
            return self._raise_error('Saving Profile: %s %s' % (pobj, ex))
        
        # Sync foreign keys    
        extensions = pobj.extensions.all()          
        extensions = nextensions + list(extensions)
        for extension in extensions:
            if extension in nextensions:
                extension.save()
                pobj.extensions.add(extension)
            else:
                pobj.extensions.remove(extension)
        
        groups = pobj.atp_groups.all()          
        groups = ngroups + list(groups)
        for group in groups:
            if group in ngroups:
                group.save()
                pobj.atp_groups.add(group)
            else:
                pobj.atp_groups.remove(group)
                
        metricinstances = pobj.metricinstances.all()          
        metricinstances = nmetricinstances + list(metricinstances)
        for metricinstance in metricinstances:
            if metricinstance in nmetricinstances:
                x,created = P_M_Instances.objects.get_or_create(profile=pobj, metricinstance=metricinstance)
                #metricinstance.save()
                #pobj.metricinstances.add(metricinstance)
            else:
                try:
                    c = P_M_Instances.objects.get(profile=pobj, metricinstance=metricinstance)
                    c.delete()
                except ObjectDoesNotExist:
                    LOG_INSTANCE.warning('Got error while deleting MetricInstance %s %s' % (pobj, metricinstance))
                #pobj.metricinstances.remove(metricinstance)
        pobj.save()
        LOG_INSTANCE.info('Synchronized profile %s' % (pobj))
        return pobj
    
    def sync_object_list(self, object_list, function_name):
        ob_list = []
        res = self._raise_exception
        self._raise_exception = False
        for ob in object_list:
            obj = function_name(ob)
            if obj:
                ob_list.append(obj)
        self._raise_exception = res
        return ob_list

    def sync_extension_list(self, ex_list):
        e_list = []
        for ex in ex_list:
            try:
                e_list.append(self.sync_extension(ex['name']))
            except:
                pass
        return e_list
    
    def urlopen_handler(self, signum, frame):
        self._raise_error("urlopen timeout occured after %s seconds." % (str(self._urlopen_timeout)))
           
    def get_data_url(self, url, append):
        if not url:
            url = self._base_url      
        try:
            if url[-1] != '/':
                url = url + '/'
            # http://stackoverflow.com/questions/5565291/detecting-hangs-with-python-urllib2-urlopen
            signal.alarm(self._urlopen_timeout)
            dcstr = simplejson.JSONDecoder().decode(urllib.urlopen(url+append).read())
            signal.alarm(0) # disable the signal
        except IOError, (er):
            self._raise_error("Error Occurred while Retrieving URL: %s%s : %s" % (url, append, er))
        except Exception, (er):
            self._raise_error("Error Occurred while JSON decode: %s" % (er))
        return dcstr
        
    def sync_metricinstance_list_url(self, url=None):
        return self.sync_object_list(self.get_data_url(url, 'metricinstances'), self.sync_metricinstance)
    
    def sync_expression_list_url(self, url=None): 
        return self.sync_object_list(self.get_data_url(url, 'expressions'), self.sync_expression)
 
    def sync_groups_list_url(self, url=None): 
        return self.sync_object_list(self.get_data_url(url, 'groups'), self.sync_groups)
    
    def sync_profile_list_url(self, url=None):
        return self.sync_object_list(self.get_data_url(url, 'profiles'), self.sync_profile)

    def get_ns_name_from_url(self, url):
        return self.get_data_url(url, 'namespace')
    
    def sync_profile_list_namespace(self, namespace):
        self.set_namespace(namespace)
        sprofile_list = self.get_selective_list()
        if sprofile_list.has_key(namespace.name):
            self.selective_profile_list = sprofile_list[namespace.name]
        else: 
            self.selective_profile_list = []
        self._profile_exist_list = []
        nprofile_list = self.sync_profile_list_url()
        profile_list = list(Profile.objects.filter(isdeleted = 'N').filter(namespace=namespace))
        for profile in profile_list:
            if profile not in nprofile_list and profile not in self._profile_exist_list:
                profile.isdeleted = 'Y'
                profile.save()
                LOG_INSTANCE.info("[%s] Profile %s has been marked Deleted" % (namespace, profile))
            if profile not in nprofile_list and profile in self._profile_exist_list:
                self._report_warning("[%s] Profile %s has degraded please check" % (namespace, profile))         
        profile_list = list(Profile.objects.filter(isdeleted = 'N'))
        return profile_list
    
    def sync_namespace(self, namespace):
        self.set_namespace(namespace)
        try:
            #self.sync_expression_list_url()
            self.sync_profile_list_namespace(namespace)
        except Exception, (er):
            self._report_error("From sync_namespace %s" % (er))
            return False
        return True
    def sync_namespace_list_from_json(self, namespace_files):
        try:
            ns_list = []
            for nfile in namespace_files :
                ns_list.extend(self.ns_sync_file(nfile))
            if ns_list :
            	nsold = Namespace.objects.all().filter(isdeleted = 'N').exclude(id__in = ns_list).update(isdeleted = 'Y')
        except Exception, (er):
            self._report_error("From sync_namespace_list_from_json %s" % (er))
            return False
        return True

    def ns_sync_file(self, file):
        nsdatas = self.get_data_url('file:///' + file, '')
        ns_list = []
        for nsdata in nsdatas :
            #nsdata = nsdata['fields']
            ns = Namespace.objects.all().filter(name = nsdata['name'])
            if not ns :
                ns = Namespace()
            else :
                ns = ns[0]
            ns.isdeleted = 'N'
            self.update_obj(ns,nsdata)
            LOG_INSTANCE.info("Added Namespace '%s'" % (ns))
            ns.save()
            ns_list.append(ns.id)
        return ns_list

    def sync_namespace_config (self):
        # two urls can't provide the same namespace, 
        # two namespaces can't be provided by same url => no two urls or namespaces can be the same
        # configured (ns, url) tuples have always preference over the database tuples 
        try: 
    	    if not Poem_sync.settings.NS_URLS:
                self._raise_error("No namespace configured please check configuration file")
    	    else:
                ns_url_list = [ self.__set_poem_api(ns_url) for ns_url in Poem_sync.settings.NS_URLS.split() ]
                ns_list = [ self.get_ns_name_from_url(ns_url)['namespace'] for ns_url in ns_url_list ]
                
                ns_dup = [k for (k, g) in groupby(sorted(ns_list)) if len(list(g)) > 1]
                url_dup = [k for (k, g) in groupby(sorted(ns_url_list)) if len(list(g)) > 1]
                
                if ns_dup or url_dup:
                    self._raise_error("Multiple identical namespaces or URLs found in the config (%s, %s)" % (str(ns_dup),
                                                                                                   str(url_dup)) )                
                ns_list = []
                for ns_url in ns_url_list:
                    ns_name = self.get_ns_name_from_url(ns_url)
                    ns_name = ns_name['namespace'] 
                    ns = Namespace.objects.filter(name=ns_name)
                    if not ns:
                        ns = Namespace()
                    else:
                        ns = ns[0]
                    ns.name = ns_name
                    ns.api_url = ns_url
                    ns.isdeleted = 'N'
                    ns.save()
                    ns_list.append(ns.id)
                if ns_list :
            	    nsold = Namespace.objects.filter(isdeleted='N').exclude(id__in=ns_list).update(isdeleted='Y')
        except Exception, (er):
            self._report_error("From sync_namespace_config %s" % (er))
            return False

    def __set_poem_api(self, ns_url):
        if 'poem/api' not in ns_url:
            return(urljoin(ns_url, Poem_sync.settings.DEFAULT_API))
        else:
            return(re.sub(r'/poem/api/\d.\d/json/', Poem_sync.settings.DEFAULT_API, ns_url))

    def get_selective_list(self):
        selective_list = {}
        if Poem_sync.settings.NS_RESTRICT:
            LOG_INSTANCE.info("Selective synchronization of profiles enabled")
            ns_pr_pairs = Poem_sync.settings.NS_RESTRICT.split()
            if ns_pr_pairs:
                try:
                    ns_pr = [ item.split('!') for item in ns_pr_pairs ]
                    for ns, pr in ns_pr:
                        selective_list.setdefault(ns, []).append(pr)                
                except Exception, (er):
                    self._report_error("Exception occurred in get_selective_list %s" % (er))

        return selective_list
    
    def unlink_namespace(self, ns_object):
        profiles2d = ns_object.profile_set.filter(isdeleted='N')
        if profiles2d:
            LOG_INSTANCE.info("Unlinking profiles %s in %s namespace", str([ p.name for p in profiles2d.all() ]),
                                                                         ns_object.name )
            profiles2d.update(isdeleted='Y')
            # Note: we should also delete related metrics and fqans here
                   
        #metric_instances = set([ profile.metric_instances for profile in profiles2d ])
        # mark as deleted all metric_instances that refer to already deleted profiles
        #mis2d = [ mi for mi in metric_instances if set(mi.profiles).issubset(profiles2d) ]
        
        