"""
POEM Synchronizer Web API is implemented using 
`Django Piston <https://bitbucket.org/jespern/django-piston/wiki/Documentation>`_ 
and follows its conventions. The implementation has two set of piston handlers, the first one 
generates REST API for POEM-SYNC models (via Django ORM):
 * :py:class:`poem_sync_api.handlers.ProfileHandler`

The second group implements custom REST resources to support bootstrapping of MRS and NCG: 
 * :py:class:`poem_sync_api.handlers.MetricInstancesHandler`
 * :py:class:`poem_sync_api.handlers.ServiceMetricInstancesHandler`

Current API is read-only as only GET method calls are supported. 

.. note::

    Please note that internal API can change without any prior notice and only 
    supports internal SAM components.

"""
from piston import handler
from taggit.utils import parse_tags

from django.http import HttpResponse, HttpResponseServerError
from django.utils import simplejson
from django.db.models import Q
from django.conf import settings

from Poem_sync.poem_sync import models 
from Poem_sync.poem_sync_api import networking

METRIC_CONFIG_MODULE = None

class ProfileHandler(handler.BaseHandler):
    """ 
    Dumps the full list of profiles with all its attributes. This is used primarily 
    for debugging purposes.
    Sample output:
    ::
      [ {
        "name": "ALICE",
        "tags": "nagios",
        "metricinstances": [
            {
                "metric": {
                    "name": "org.alice.vobox-DPD",
                    "isdeleted": "N"
                }, ... ]
        }
      } ]
       
    * *URL*: ``/api/0.1/<format>/profiles/`` 
    * *Supported formats*: xml, json
    * *Supported methods*: GET
    """
    model = models.Profile
    allowed_methods = ('GET')
    fields = ('id', 'name', 'version',
              'valid_from', 'valid_to', 'tags', 'owner', 'description', 'isdeleted',
               ('atp_vo', ('voname','isdeleted')),
               ('namespace', ('name','api_url', 'isdeleted')), 
               ('atp_groups', ('groupname', ('group_type', ('typename',)),'isdeleted')),
               ('metricinstances', (
                    ('metric', ('name','isdeleted')), 
                    ('fqan', ('fqan','isdeleted')), 
                    ('atp_service_type_flavour', ('flavourname','isdeleted')),
                    ('atp_vo', ('voname','isdeleted'))
               )),
               ('extensions', (
                    ('expression', ('expr','name', 'expr_type')),
                ))
             )

    def read(self, request, *args, **kwargs):
        filters = dict([ (k.encode('ascii'), v.encode('ascii')) for k, v in request.GET.items() ])
        if not 'isdeleted' in filters.keys(): # don't show deleted profiles
            filters['isdeleted']='N'
        return super(ProfileHandler, self).read(request, *args, **filters)

class MetricInstancesHandler(handler.BaseHandler):
    """ 
    Lists all unique metric instances for all profiles (if POEM_NAGIOS_PROFILES exists then
    only profiles in this variable are considered). Metric instances are
    tuples of (flavour, metric, vo, fqan). This is used by NCG to bootstrap Nagios. Entries
    with no FQAN are interpreted as pure VO metric instances. 
    The listed metrics needs to have a working configuration stored in /etc/ncg-metric-config.conf
    or /etc/ncg-metric-config.d/*.conf
    Sample output:
    ::
        [ [ "org.atlas.WN-swtag", "CREAM-CE", "atlas", "/atlas/Role=lcgadmin" ], ... ]
       
    * *URL*: ``/api/0.1/<format>/metricinstances/`` 
    * *Supported formats*: json
    * *Supported methods*: GET
    """    
    def read(self, request, *args, **kwargs):
        profile_filter = None
        filters = dict([ (k.encode('ascii'), v.encode('ascii')) for k, v in request.GET.items() ])
        if 'profile' in filters.keys():
            profile_filter = filters['profile']
            del(filters['profile'])
        
        try:
            mis = models.MetricInstance.objects.filter(profiles__isdeleted='N')
            
            # filtering based on nagios tag was removed (waiting for proper solution)
            # pt = models.Profile.objects.filter(isdeleted='N').values_list('name', 'tags')
            # pt_split = [ (item[0], parse_tags(item[1])) for item in pt ]
            # tagged_pts = [ item[0] for item in pt_split if 'nagios' in item[1] ]
            #if tagged_pts:
            #    mis = mis.filter(profiles__name__in=tagged_pts)
            # backward compatibility in case tags are completely missing (can be removed in the future)
            if settings.POEM_NAGIOS_PROFILES:
                top_level_filter = settings.POEM_NAGIOS_PROFILES.split()
                mis = mis.filter(profiles__name__in=top_level_filter)

            if profile_filter:
                # functionality is limited here as you can have multiple 
                # profiles with same name in diff. namespaces (for Site-Nagios should be ok)
                mis = mis.filter(profiles__name=profile_filter)
                                
            mis2 = mis.filter(Q(fqan__fqan__isnull=True) | Q(fqan__isdeleted='N'))
            if filters:
                mis2 = mis2.filter(**filters)

            mis_values = list(set(mis2.values_list('metric__name', 'atp_service_type_flavour__flavourname',
                                     'atp_vo__voname', 'fqan__fqan' )))
            return HttpResponse(simplejson.JSONEncoder().encode(mis_values),
                                mimetype='application/json')
        except Exception, e:
            return HttpResponseServerError('Exception occurred while trying to generate service metric instances (%s).' % str(e))
        
            
class ServiceMetricInstancesHandler(handler.BaseHandler):
    """ 
    Lists all unique service metric instances for all the profiles. Service metric instances are
    tuples of (service, flavour, metric, vo, fqan). This is used to bootstrap MRS. Entries
    with no FQAN will be interpreted by MRS as accepting any FQAN. In addition, post-processing is 
    performed to mark VO independent metrics with null FQAN and VO.
    The listed metrics needs to have a working configuration stored in /etc/ncg-metric-config.conf
    or /etc/ncg-metric-config.d/*.conf. Sample output:
    ::
        [ [ "ce-iep-grid.saske.sk", "org.atlas.WN-swtag", "CREAM-CE", "atlas", "/atlas/Role=lcgadmin" ], ... ]
       
    * *URL*: ``/api/0.1/<format>/metricinstances/`` 
    * *Supported formats*: json
    * *Supported methods*: GET
    """       
    def read(self, request, *args, **kwargs):
        try:
            # assuming at least one profile exists for each VO without specifying any atp group
            vos = reduce(lambda x, y: x+y, set(models.Profile.objects.filter(isdeleted='N').values_list('atp_vo__voname')))
            total_smi = list()
        
            for vo in vos:
                status, srvs = networking.get_services(vo)
                if status:
                    # if ATP gives empty set for any VO then fail
                    # this is to prevent MRS from wiping out entire VO
                    if not srvs:
                        return HttpResponseServerError('No services for %s from ATP, bailing out.' % str(vo))
                    # all non-deleted profiles
                    mis = models.MetricInstance.objects.filter(atp_vo__voname__exact=vo,
                                                               profiles__isdeleted='N')
                    mis = mis.filter(Q(fqan__isdeleted='N') | Q(fqan__fqan__isnull=True)).values_list(
                                                    'metric', 'metric__name',
                                                    'atp_service_type_flavour', 
                                                    'atp_service_type_flavour__flavourname',
                                                    'fqan','fqan__fqan', 'atp_vo', 'atp_vo__voname')
                    # join based on service flavour ids
                    vo_smi = [ (s+(m[0],m[1])+m[4:]) for s in srvs for m in mis if s[2]==m[2] ]
                    
                    # Fix vo independent metrics with lazy loading of metric_config module
                    global METRIC_CONFIG_MODULE
                    if not METRIC_CONFIG_MODULE:
                        from Poem_sync.poem_sync_api import metric_config
                    vo_indep_metrics =  metric_config.get_vo_indep_metrics()
                    # from python 2.5 onwards map(lambda x: x if x[5] not in y else x[:6]+None..., vo_smi)
                    vo_smi = map(lambda entry: entry[5] in vo_indep_metrics and entry[:6]+(None, None, None, None) or entry, vo_smi)
                    
                    total_smi = total_smi + vo_smi
                else:
                    return HttpResponseServerError('Failed to get services for %s from ATP, bailing out.' % str(vo))
            
            total_smi = list(set(total_smi))
            return HttpResponse(simplejson.JSONEncoder().encode(total_smi),
                                mimetype='application/json')
                
        except Exception, e:
            return HttpResponseServerError('Exception occurred while trying to generate service metric instances (%s).' % str(e))
    